/*
  Copyright © 2021 Jan Bobrowski

  This program is free software; you can redistribute it or
  modify it under the terms of the GNU General Public License
  version 2 as published by the Free Software Foundation.

  Special thanks to Jason Summers for BMP Suite
  http://entropymine.com/jason/bmpsuite/
*/

#include <strings.h>
#include "toico.h"
#include "rec.h"

static inline u32 get24le(void *b) {u8 *p = b; return p[0] | p[1]<<8 | p[2]<<16;}

int is_bmp(u8 p[SIGN_SIZE])
{
	return p[0] == 0x42 && p[1] == 0x4d && (SIGN_SIZE < 14 || p[12] == 0 && p[13] == 0);
}

struct bitfield {
	u32 m;
	u8 r, l, w, w2, w3;
};

static void bf_init(struct bitfield *bf, u32 m);
static u32 bf_map(struct bitfield *bf, u32 v);
static void read_rle(struct image *im, struct file *fi, int compr);

void read_bmp(struct image *im, struct file *fi)
{
	union {
		struct {
			struct BMP_Header h;
			union {
				u32le size;
				struct BitmapCoreHeader c;
				struct BitmapInfoHeader i;
				struct BitmapV4Header v4;
				struct BitmapV5Header v5;
				struct BitmapCoreHeader2 c2;
			};
		};
		u8 b[sizeof(struct BMP_Header) + sizeof(struct BitmapV5Header)];
	} r;
	u32 ofs, size, compr, cu;
	int depth;
	int t_b = 0;
	struct bitfield bfr, bfg, bfb, bfa;
	u32le *bf_ptr = 0;
	u32le *bfa_ptr = 0;

	{
		u32 have;
		u32 need = sizeof r.h + 4;
		read_all(fi, r.b, need);
		have = need;
		size = get32le(r.size);
		if (size > 0xffff)
			bad("Format not supported");
		need = sizeof r.h + size;
		if (need > sizeof r.b)
			bad("Format not supported");
		if (have < need) {
			read_all(fi, r.b + have, need - have);
			have = need;
		}
	}

	switch (size) {
		u32 h;
	case sizeof r.c:
		cu = 0;
		if (get16le(r.c.bcPlanes) > 1)
			bad("Format not supported");
		im->width = get16le(r.c.bcWidth);
		h = get16le(r.c.bcHeight);
		if (h & 0x8000) {
			h = 0x10000 - h;
			t_b = 1;
		}
		im->height = h;
		depth = get16le(r.c.bcBitCount);
		compr = Compr_None;
		break;
	case sizeof r.i:
	case sizeof r.v4:
	case sizeof r.v5:
	case sizeof r.c2:
		cu = get32le(r.i.biClrUsed);
		depth = get16le(r.i.biBitCount);
		if (cu > 0x10000 || get16le(r.i.biPlanes) > 1)
			bad("Format not supported");
		compr = get32le(r.i.biCompression);
		switch (compr) {
		case Compr_None:
			if (depth == 16) {
				static u32le masks[3] = {{0x00, 0x7c}, {0xe0, 0x03}, {0x1f, 0x00}};
				bf_ptr = masks;
				compr = Compr_Bitfields;
			}
			break;
		case Compr_Bitfields:
		case Compr_AlphaBitfields:
			if (size == sizeof r.v4 || size == sizeof r.v5) {
				bf_ptr = &r.v4.bV4RedMask;
				bfa_ptr = &r.v4.bV4AlphaMask;
			}
			break;
		case Compr_RLE4:
			if (depth != 4)
				bad("Bad depth for RLE (%d)", depth);
			break;
		case Compr_RLE8:
			if (depth != 8)
				bad("Bad depth for RLE (%d)", depth);
			break;
		default:
			bad("Compression not supported (%08X)", compr);
		}
		h = get32le(r.i.biHeight);
		if (h & 0x80000000u) {
			h = -h;
			t_b = 1;
		}
		im->height = h;
		im->width = get32le(r.i.biWidth);
		break;
	default:
		bad("BMP not recognized");
	}

	ofs = sizeof r.h + size;

	if (compr == Compr_Bitfields || compr == Compr_AlphaBitfields) {
		u32le b[4];
		u32 rm, gm, bm, am;
		int sz = compr == Compr_Bitfields ? 3*4 : 4*4;
		if (!bf_ptr) {
			ofs += sz;
			read_all(fi, b, sz);
			bf_ptr = b;
			if (sz == 16)
				bfa_ptr = b + 3;
		}
		rm = get32le(bf_ptr[0]);
		gm = get32le(bf_ptr[1]);
		bm = get32le(bf_ptr[2]);
		if (depth == 32 && !bfa_ptr && rm == 0xff0000 && gm == 0x00ff00 && bm == 0x0000ff) {
			compr = Compr_None;
		} else {
			bf_init(&bfr, rm);
			bf_init(&bfg, gm);
			bf_init(&bfb, bm);
			if (bfa_ptr) {
				am = get32le(*bfa_ptr);
				compr = am ? Compr_AlphaBitfields : Compr_Bitfields;
				if (am) bf_init(&bfa, am);
			}
		}
	}

	if (!cu && depth <= 8)
		cu = 1 << depth;
	if (cu) {
		int s = size == sizeof r.c ? 3 : 4;
		u8 *b = alloc(cu*s);
		int i;
		ofs += cu*s;
		read_all(fi, b, cu*s);
		if (cu > elemof(im->pal))
			cu = elemof(im->pal);
		for (i=0; i<cu; i++)
			im->pal[i] = get24le(b + s*i);
		free(b);
	}

	{
		u32 hofs =  get32le(r.h.bfOffBits);
		skip(fi, hofs - ofs);
		ofs = hofs;
	}

	alloc_im(im);

	if (compr == Compr_RLE4 || compr == Compr_RLE8)
		read_rle(im, fi, compr);
	else {
		u8 *b = alloc(roundup(im->width, depth));
		u32 *e = t_b ? im->data + im->width : data_end(im);
		for (;;) {
			read_all(fi, b, roundup(im->width, depth));
			u32 *p = e - im->width;
			u8 *q = b;
			if (compr == Compr_None) switch (depth) {
			case 1:
			case 2:
			case 4:
			case 8:
				for (;;) {
					u32 v = get32be_at(q);
					u32 m = (1 << depth) - 1;
					int s = 32;
					q += 4;
					do {
						s -= depth;
						*p++ = im->pal[v>>s & m];
						if (p == e)
							goto line_done;
					} while (s);
				}
			case 24:
			case 32:
				do {
					u32 v = get24le(q);
					q += depth >> 3;
					*p++ = v;
				} while (p < e);
				break;
			default:
				bad("Depth not supported (%u)", depth);
			} else {
				if (depth != 16 && depth != 32)
					bad("Bad depth (%u)", depth);
				do {
					u32 v = q[0] | q[1] << 8;
					q += 2;
					if (depth == 32) {
						v |= (q[0] | q[1] << 8) << 16;
						q += 2;
					}
					*p++ = bf_map(&bfr, v)<<16 | bf_map(&bfg, v)<<8 | bf_map(&bfb, v);
					if (compr == Compr_AlphaBitfields)
						p[-1] |= (bf_map(&bfa, v) << 24) ^ ALPHA;
				} while (p < e);
			}
line_done:
			if (t_b) {
				e += im->width;
				if (e > data_end(im))
					break;
			} else {
				e -= im->width;
				if (e <= im->data)
					break;
			}
		}
		free(b);
	}
}

static void bf_init(struct bitfield *bf, u32 m)
{
	int s, w;
	*bf = (struct bitfield){};
	if (!m)
		return;
	s = ffs(m) - 1;
	w = ffs((m >> s) + 1) - 1;
	bf->m = m;
	bf->w = w;
	s += w - 8;
	if (s < 0) bf->l = -s;
	else bf->r = s;
	w += w;
	if (w < 8) {
		bf->w2 = w;
		w += w;
		if (w < 8)
			bf->w3 = w;
	}
}

static u32 bf_map(struct bitfield *bf, u32 v)
{
	v = (v & bf->m) >> bf->r << bf->l;
	v |= v >> bf->w;
	v |= v >> bf->w2;
	v |= v >> bf->w3;
	return v & 0xff;
}

static void read_rle(struct image *im, struct file *fi, int compr)
{
	u32 *e = data_end(im);
	int skip_lines = 0;
	do {
		u32 *p = e - im->width;

		if (skip_lines) {
			if (skip_lines > 0)
				skip_lines--;
			goto skip_rest;
		}

		for (;;) {
			u8 b[2];
			read_all(fi, b, 2);
			if (b[0]) {
				int n = b[0];
				if (n > e - p) {
					n = e - p;
					if (!n) continue;
				}
				if (compr == Compr_RLE4) {
					u32 c0 = im->pal[b[1] >> 4];
					u32 c1 = im->pal[b[1] & 0xf];
					do {
						*p++ = c0;
						if (!--n) break;
						*p++ = c1;
					} while (--n);
				} else {
					u32 c = im->pal[b[1]];
					do {
						*p++ = c;
					} while (--n);
				}
			} else switch (b[1]) {
				int n;
			case 1:
				skip_lines = -1;
			case 0:
				goto skip_rest;
			case 2:
				read_all(fi, b, 2);
				skip_lines = b[1];
				n = b[0];
				if (p + n >= e)
					goto skip_rest;
				while (n--)
					*p++ = ALPHA;
				break;
			default:
				n = b[1];
				if (p + n > e) bad("Bad RLE");
				do {
					read_all(fi, b, 2);
					if (compr == Compr_RLE4) {
						*p++ = im->pal[b[0]>>4 & 0xf];
						if (!--n) break;
						*p++ = im->pal[b[0] & 0xf];
						if (!--n) break;
						*p++ = im->pal[b[1]>>4 & 0xf];
						if (!--n) break;
						*p++ = im->pal[b[1] & 0xf];
					} else {
						*p++ = im->pal[b[0]];
						if (!--n) break;
						*p++ = im->pal[b[1]];
					}
				} while (--n > 0);
			}
		}
skip_rest:
		while (p < e) {
			*p++ = ALPHA;
		}
		e -= im->width;
	} while (e > im->data);
}
