/*
  Copyright © 2001-2022 Jan Bobrowski

  This program is free software; you can redistribute it or
  modify it under the terms of the GNU General Public License
  version 2 as published by the Free Software Foundation.
*/

#include "toico.h"
#include "rec.h"

int is_ico(u8 sig[SIGN_SIZE])
{
	static u8 begining[] = {0, 0, 1, 0};
	return memcmp(sig, begining, sizeof begining) == 0 && (sig[4] | sig[5]<<8) != 0;
}

static void set16le(u16le *p, u16 v) {p->a = v; p->b = v >> 8;}
static void put24(u8 p[3], u32 v) {p[0] = v; p[1] = v >> 8; p[2] = v >> 16;}
static void set32le(u32le *p, u32 v) {p->a = v; p->b = v >> 8; p->c = v >> 16; p->d = v >> 24;}

static void fwrite_all(void *buf, int n, FILE *f)
{
	char *p = buf;
	do {
		int v = fwrite(p, 1, n, f);
		if (!v) err(1, "fwrite");
		p += v; n -= v;
	} while(n);
}

static void fput32be(u32 v, FILE *f)
{
	u8 b[4] = {v >> 24, v >> 16, v >> 8, v};
	fwrite_all(b, 4, f);
}

static inline int findidx(struct image *i, u32 c)
{
	int n = 0;
	if (transparent(c))
		return 0;
	for (;;) {
		if (i->pal[n] == c)
			return n;
		n++;
	}
}

static void outbmap(struct image *im, FILE *f)
{
	u32 *e = data_end(im);
	int d = im->depth;
	do {
		u32 *p = e - im->width;
		do {
			int s = 32;
			u32 b = 0;
			do {
				u32 c = findidx(im, *p++);
				s -= d;
				b |= c<<s;
			} while (s > 0 && p < e);
			fput32be(b, f);
		} while (p < e);
		e -= im->width;
	} while (e > im->data);
}

static void outbmap24(struct image *im, FILE *f)
{
	u32 *e = data_end(im);
	do {
		u32 *p = e - im->width;
		do {
			u8 b[3];
			put24(b, *p++);
			fwrite_all(b, 3, f);
		} while (p < e);
		if (im->width & 3) {
			static u8 z[3] = {0, 0, 0};
			fwrite_all(z, im->width & 3, f);
		}
		e -= im->width;
	} while (e > im->data);
}

static void outbmap32(struct image *im, FILE *f)
{
	u32 *e = data_end(im);
	do {
		u32 *p = e - im->width;
		do {
			u32le b;
			set32le(&b, *p++ ^ ALPHA);
			fwrite_all(&b, 4, f);
		} while (p < e);
		e -= im->width;
	} while (e > im->data);
}

static void outmask(struct image *im, FILE *f)
{
	u32 *e = data_end(im);
	do {
		u32 *p = e - im->width;
		do {
			int s=32; u32 b=0;
			do {
				s--;
				if (p >= e)
					break;
				if (*p++ >= 0xF0000000)
					b |= 1<<s;
			} while(s > 0);
			fput32be(b, f);
		} while (p < e);
		e -= im->width;
	} while (e > im->data);
}

static void writepal(struct image *im, FILE *f)
{
	u32le buf[256];
	u32 *p = im->pal;
	int i = im->npal;
	while (i-- > 0)
		set32le(&buf[i], p[i]);
	fwrite_all(buf, 4*im->npal, f);
}

/*
 * Checks presence of full alpha channel.
 * Also sets transparent color to black.
 */
static int fullalpha(struct image *i)
{
	u32 *p = i->data;
	u32 *e = data_end(i);
	u32 y = ALPHA;
	do {
		u32 v = *p & ALPHA;
		if (v) {
			y &= v;
			if (v == ALPHA)
				*p = v;
		}
	} while (++p < e);
	return y != ALPHA;
}

/*
 * Truncates alpha channel to 1-bit (mask).
 * Also sets transparent color to black.
 */
static void alphatr(struct image *im, u8 threshold)
{
#define BGLEVEL 0xD0
	u32 *p = im->data, *e = data_end(im);
	u32 thr = threshold << 24;
	do {
		u32 v = *p++;
		if (v >= thr)
			p[-1] = ALPHA;
		else if (v & ALPHA) {
			int a = v>>24 & 0xFF;
			u8 r = v>>16 & 0xFF;
			u8 g = v>>8 & 0xFF;
			u8 b = v & 0xFF;
			v &= ~ALPHA;
			if (4*b + 2*r + b <= 7*BGLEVEL) {
				int ca = a ^ 0xFF;
				r = (r*ca + BGLEVEL*a + 0x7F) / 0xFF;
				g = (g*ca + BGLEVEL*a + 0x7F) / 0xFF;
				b = (b*ca + BGLEVEL*a + 0x7F) / 0xFF;
				v = r<<16 | g<<8 | b;
			}
			p[-1] = v;
		}
	} while (p < e);
}

static int index_im(struct image *im)
{
	u32 *e = data_end(im);
	u32 *p;
	int n = 0;

	p = im->data;
	do if (transparent(*p++)) {
		im->pal[n++] = 0;
		break;
	} while (p < e);

	p = im->data;
	do {
		u32 c = *p++;
		int i;
		if (transparent(c))
			continue;
		for (i = 0; i < n; i++)
			if (im->pal[i] == c)
				goto next;
		if (n == 256)
			return 0;
		im->pal[n++] = c;
next:;
	} while (p < e);

	if (n <= 16) {
		im->depth =  4;
		if (n <= 2)
			im->depth = 1;
	} else {
		unsigned a = 4*n + im->height*roundup(im->width, 8);
		unsigned b = im->height*roundup(im->width, im->depth);
		if (b <= a)
			return 0;
		im->depth = 8;
	}
	im->npal = n;
	return 1;
}

void write_ico(FILE *f)
{
	struct list *l;
	int ofs;

	{
		struct ICO_Header h;
		set16le(&h.idReserved, 0);
		set16le(&h.idType, 1);
		set16le(&h.idCount, nimages);
		fwrite_all(&h, sizeof h, f);
	}

	ofs = sizeof(struct ICO_Header) + nimages*sizeof(struct IcoDirEntry);

	for (l = images.next; l != &images; l = l->next) {
		struct image *im = list_item(l, struct image, list);
		struct IcoDirEntry e = {};
		int sz, cc;

		cc = 0;
		if (!im->embed) {
			if (!im->no_semi) {
				if (!fullalpha(im))
					goto index;
				im->depth = 32;
			} else {
				alphatr(im, 127);
index:
				im->depth = 24;
				index_im(im);
			}

			cc = im->npal;
			sz = sizeof(struct BitmapInfoHeader);
			sz += 4*cc + im->height*(roundup(im->width, im->depth) + roundup(im->width, 1));
		} else {
			assert(im->raw_data);
			sz = im->raw_size;
//			sz = roundup(im->raw_size, 8);
		}

		if (opt.verbose) {
			fprintf(stderr,"%-16s %dx%d", im->name, im->width, im->height);
			if (im->depth)
				fprintf(stderr,"x%d", im->depth);
			if (im->npal)
				fprintf(stderr, " colors:%d", im->npal);
			putc('\n', stderr);
		}

		e.bWidth = im->width > 255 ? 0 : im->width;
		e.bHeight = im->height > 255 ? 0 : im->height;
		e.bColorCount = cc;
		set16le(&e.wPlanes, 1);
		set16le(&e.wBitCount, im->depth);
		set32le(&e.dwBytesInRes, sz);
		set32le(&e.dwImageOffset, ofs);
		fwrite_all(&e, sizeof e, f);
		ofs += sz;
	}

	for (l = images.next; l != &images; l = l->next) {
		struct image *im = list_item(l, struct image, list);
		struct BitmapInfoHeader b = {};

		if (im->embed) {
			fwrite_all(im->raw_data, im->raw_size, f);
			// XXX align?
			continue;
		}

		set32le(&b.biSize, sizeof b);
		set32le(&b.biWidth, im->width);
		set32le(&b.biHeight, 2*im->height);
		set16le(&b.biPlanes, 1);
		set16le(&b.biBitCount, im->depth);
		set32le(&b.biClrUsed, im->npal);
		fwrite_all(&b, sizeof b, f);

		switch (im->depth) {
		case 1:
		case 4:
		case 8:
			writepal(im, f);
			outbmap(im, f);
			break;
		case 24:
			outbmap24(im, f);
			break;
		case 32:
			outbmap32(im, f);
		}
		outmask(im, f);
	}
}


union bmih {
	u32le size;
	struct BitmapInfoHeader ih;
};

void list_ico(struct file *fi)
{
	struct ICO_Header h;
	struct IcoDirEntry *ents;
	union bmih *bmhs;
	u32 ofs;
	int i, n;

	read_all(fi, &h, sizeof h);
	n = get16le(h.idCount);
	ents = alloc(n * (sizeof *ents + sizeof *bmhs));
	bmhs = (union bmih*)(ents + n);
	read_all(fi, ents, n * sizeof *ents);
	ofs = sizeof h + n * sizeof *ents;

	for (;;) {
		union bmih *bmh = 0;
		u32 mio;

		for (i=0; i<n; i++) {
			u32 o = get32le(ents[i].dwImageOffset);
			if (ofs == o) {
				if (!bmh) {
					bmh = &bmhs[i];
					read_all(fi, &bmhs[i], sizeof bmhs[i]);
					ofs += sizeof bmhs[i];
				} else {
					memcpy(&bmhs[i], bmh, sizeof bmhs[i]);
				}
			}
		}

		mio = ~0u;
		for (i=0; i<n; i++) {
			u32 o = get32le(ents[i].dwImageOffset);
			if (ofs < o && o < mio)
				mio = o;
		}

		if (mio == ~0u)
			break;

		skip(fi, mio - ofs);
		ofs = mio;
	}

	for (i=0; i<n; i++) {
		struct IcoDirEntry *e = &ents[i];
		union bmih *bh = &bmhs[i];
		enum {Unknown, DIB, PNG} type;

		printf("%2u. ", i);

		type = Unknown;
		if (is_png((u8*)bh)) {
			type = PNG;
		} else if (get32le(bh->size) >= sizeof bh->ih) {
			type = DIB;
		}

		if (!opt.verbose) {
			int width = e->bWidth;
			int height = e->bHeight;
			int depth = get16le(e->wBitCount);

			if (!width || !height) {
				if (type == DIB) {
					int v = get32le(bh->ih.biHeight);
					if (v < 0)
						v = -v;
					if (v & 1)
						type = Unknown;
					else {
						height = v >> 1;
						width = get32le(bh->ih.biWidth);
					}
				} else if (type == PNG)
					png_dims((u8*)bh, &width, &height);
			}

			int w, npal;
			w = 10;
			w -= printf("%ux%u", width, height);
			if (depth) w -= printf("x%u", depth);
			printf("%*s", w < 1 ? 1 : w, "");
			npal = e->bColorCount;
			if (!npal && depth > 0 && depth <= 8)
				npal = 1 << depth;
			if (npal)
				printf(" colors:%d", npal);
			if (type == PNG)
				printf(" PNG");
		} else {
			static int widths[] = {11, 3, 11, 3, 11, 3, 0};
//			static char *cns[] = {"dim", "col", "position", "typ", "dim", "col", "compr"};
			static char* compr[] = {"RGB"};
			int nk, k;

			nk = type == DIB ? elemof(widths) : 3 + (type == PNG);
			for (k = 0;;) {
				int w = widths[k];
				switch (k) {
					char *s;
					unsigned v;
				case 0:
					w -= printf("%ux%ux%u", e->bWidth, e->bHeight, get16le(e->wBitCount));
					break;
				case 1:
					w -= printf("%3u", e->bColorCount);
					break;
				case 2:
					w -= printf("%05X:%05X", get32le(e->dwImageOffset), get32le(e->dwBytesInRes));
					break;
				case 3:
					w -= printf("%s", type == DIB ? "DIB" : "PNG");
					if (type == PNG) {
						int width, height;
						if (png_dims((u8*)bh, &width, &height) == 0)
							w -= printf(" %ux%u", width, height);
					}
					break;
				case 4:
					w -= printf("%ux%ux%u", get32le(bh->ih.biWidth), get32le(bh->ih.biHeight), get16le(bh->ih.biBitCount));
					break;
				case 5:
					w -= printf("%3u", get32le(bh->ih.biClrUsed));
					break;
				case 6:
					v = get32le(bh->ih.biCompression);
					s = v < elemof(compr) ? compr[v] : 0;
					if (s)
						w -= printf("%s", s);
					else
						w -= printf("%u", v);
					break;
				}
				if (++k == nk)
					break;
				printf("%*s", w < 1 ? 1 : w, "");
			}
		}

		printf("\n");
	}
	free(ents);
}
