/*
  Copyright © 2001-2022 Jan Bobrowski

  This program is free software; you can redistribute it or
  modify it under the terms of the GNU General Public License
  version 2 as published by the Free Software Foundation.

  torinak.com/~jb/toico/
*/

#include <unistd.h>
#include <errno.h>
#include <stdarg.h>
#include <unistd.h>
#include "toico.h"

struct opt opt;

int nimages = 0;
struct list images;
struct image *current;

void bad_image(char *fmt, ...)
{
	va_list a;
	va_start(a, fmt);
	verrx(1, fmt, a);
}

void *alloc(unsigned size)
{
	void *p = malloc(size);
	if (!p) err(1, "%s", current ? current->name : "?");
	return p;
}

static void open_file(struct file *fi, char *path)
{
	fi->file = stdin;
	fi->name = "<stdin>";
	fi->ptr = fi->end = fi->buf;

	if (path && strcmp(path, "-") != 0) {
		fi->name = path;
		fi->file = fopen(fi->name, "rb");
		if (!fi->file) err(1, "%s", fi->name);
	}
}

void *read_checked(struct file *fi, void *buf, int n, int all)
{
	u8 *p = buf;
	if (n == 0)
		return p;
	while (fi->ptr < fi->end) {
		*p++ = *fi->ptr++;
		if (!--n)
			return p;
	}
	do {
		int v = fread(p, 1, n, fi->file);
		if (!v)
			break;
		p += v; n -= v;
	} while (n);
	if (ferror(fi->file))
		errx(1, "%s: Error", fi->name);
	if (n && all == All)
		errx(1, "%s: Unexpected end of file", fi->name);
	return p;
}

static u8 *peek_signature(struct file *fi) {
	read_all(fi, fi->buf, sizeof fi->buf);
	fi->ptr = fi->buf;
	fi->end = fi->buf + sizeof fi->buf;
	return fi->buf;
}

void skip(struct file *fi, u32 n)
{
	u32 k = fi->end - fi->ptr;
	if (n < k)
		k = n;
	fi->ptr += k;
	n -= k;
	if (!n)
		return;
	if (fseek(fi->file, n, SEEK_CUR) == 0)
		return;
	if (errno != ESPIPE)
		err(1, "%s", fi->name);
	do {
		if (fgetc(fi->file) == EOF)
			errx(1, "%s: Truncated", fi->name);
	} while (--n);
}

void alloc_im(struct image *im)
{
	im->data = (u32*)alloc(im->width * im->height * sizeof(*im->data));
}

struct type_info {
	unsigned multiple:1;
	int (*is)(u8[]);
	void (*read)(struct image *, struct file *);
	void (*list)(struct file *);
};

static struct type_info types[] = {
#ifdef WITH_PNG
	{0, is_png, read_png},
#endif
	{0, is_gif, read_gif},
#ifdef WITH_TIFF
	{0, is_tiff, read_tiff},
#endif
	{0, is_bmp, read_bmp},
	{0, is_xpm, read_xpm},
	{1, is_icns, read_icns, list_icns},
	{1, is_ico, 0, list_ico}
};

static struct type_info *identify_type(u8 sig[SIGN_SIZE])
{
	int i = 0;
	while (!types[i].is(sig)) {
		if (++i == elemof(types))
			return 0;
	}
	return &types[i];
}

static void read_im(struct image *im)
{
	u8 *sig;
	struct file fi;
	struct type_info *ti;

	open_file(&fi, im->name);
	current = im;

	if (im->embed) {
		unsigned have = 16 << 10;
		u8 *buf = alloc(have);
		unsigned loaded = 0;
		for (;;) {
			loaded = (u8*)read_some(&fi, buf + loaded, have - loaded) - buf;
			if (loaded < have)
				break;
			have *= 2;
			buf = realloc(buf, have);
			if (!buf)
				err(1, "%s", fi.name);
		}
		if (!is_png(buf) || png_dims(buf, &im->width, &im->height) < 0)
			errx(1, "%s: Not a PNG", fi.name);
		im->raw_size = loaded;
		im->raw_data = realloc(buf, loaded);
		goto close_ret;
	}

	sig = peek_signature(&fi);

	ti = identify_type(sig);
	if (!ti)
		errx(1, "%s: Unsupported file format", fi.name);
	if (!ti->read)
		errx(1, "%s: Reading not supported", fi.name);
	if (ti->multiple ^ (im->index >= 0)) {
		if (!ti->multiple)
			errx(1, "%s: Not a collection", fi.name);
		im->index = 0;
	}
	ti->read(im, &fi);

close_ret:
	if (fi.file != stdin)
		fclose(fi.file);
	current = 0;
}

static void add_transparency(struct image *im)
{
	u32 *p = im->data;
	u32 *e = data_end(im);
	do {
		if (*p == im->alpha_key)
			*p = ALPHA;
	} while (++p < e);
}

static _Noreturn void usage(int y)
{
	if (opt.verbose) {
		printf(STR(VERSION) "\n");
		exit(0);
	}
	if (y)
		printf("ToICO " STR(VERSION) "; © 2001-2022 Jan Bobrowski; GPL\n\n");
	printf(
		"toico [-v] img1-opts img1 [img2-opts img2]... [-o icon]\n"
		"toico [-l] [icon]\n"
		"toico -h\n"
	);
	if (y) printf(
		" -o  Output to a file instead of stdout\n"
		" -l  List icon\n"
		" -v  Verbose\n"
		"img opts:\n"
		" -e  Embed raw PNG\n"
		" -1  Force 1-bit transparency\n"
		" -i num  Image index in file\n"
		" -k RRGGBB  Transparency key (eg. magenta: FF00FF)\n"
		"Reads:"
#ifdef WITH_PNG
		" PNG"
#endif
		" GIF"
#ifdef WITH_TIFF
		" TIFF"
#endif
		" BMP XPM ICNS\n\n"
		"torinak.com/~jb/toico/\n"
	);
	exit(!y);
}

static struct image *get_current()
{
	if (!current) {
		current = alloc(sizeof *current);
		*current = (struct image){.index = -1};
		list_init(&current->list);
	}
	return current;
}

struct opt_state {
	char **ptr;
	char *opt;
};

static char *opt_arg(struct opt_state *os)
{
	char *p = os->opt;
	if (p) {
		os->opt = 0;
		os->ptr++;
	} else {
		p = *os->ptr;
		if (p)
			os->ptr++;
	}
	return p;
}

static void parse_args(char **argv)
{
	struct opt_state os = {.ptr = argv};

	do {
		int o = 0;
		if (!os.opt) {
			char *p = *os.ptr;
			if (!p || p[0] != '-' || !p[1])
				goto after_opt;
			if (p[1] == '-') {
				if (!p[2])
					os.ptr++;
				goto after_opt;
			}
			os.opt = p + 1;
		}
		o = *os.opt++;
		if (!*os.opt) {
			os.opt = 0;
			os.ptr++;
		}

		switch (o) {
			case 'o':
				opt.mode_convert = 1;
				opt.file = opt_arg(&os);
				break;
			case 'l':
				opt.mode_list = 1;
				break;
			case 'x':
			case 'd':
			case '1': get_current()->no_semi = o != 'x'; break;
			case 'e': get_current()->embed = 1; break;
			case 'k': {
				char *p = opt_arg(&os);
				char *q;
				if (!p)
					errx(2, "Expected color");
				get_current()->alpha_key = strtoul(p, &q, 16);
				if (*q)
					errx(2, "`%s' should be in hex format", p);
				current->has_key = 1;
			} break;
			case 'i': {
				char *arg = opt_arg(&os);
				char *end;
				long v = strtol(arg, &end, 0);
				if (v < 0 || !*arg || *end)
					errx(2, "Bad value: %s", arg);
				get_current()->index = v;
			} break;
			case 'v': opt.verbose = 1; break;
			case 'h': usage(1);
			default:
				errx(2, "Unknown option: -%c", o);
		}
		continue;

after_opt:
		if (opt.mode_list) {
			if (opt.mode_convert)
				errx(2, "Either list or convert");
			opt.file = *os.ptr;
			if (opt.file && os.ptr[1])
				errx(2, "Too many arguments");
			return;
		}
		if (!*os.ptr) {
			if (current)
				errx(2, "No image");
			break;
		}
		get_current()->name = *os.ptr++;
		list_append(&images, &current->list);
		nimages++;
		current = 0;
		opt.mode_convert = 1;
	} while (os.ptr);
}

int main(int argc, char *argv[])
{
	struct list *l;
	FILE *out;

	if (argc < 2)
		usage(0);

	list_init(&images);
	parse_args(argv + 1);

	if (opt.mode_list) {
		struct file fi;
		struct type_info *ti;
		u8 *sig;

		open_file(&fi, opt.file);
		sig = peek_signature(&fi);
		ti = identify_type(sig);
		if (!ti)
			errx(1, "%s: Unsupported file format", fi.name);
		if (!ti->list)
			errx(1, "%s: Listing not supported", fi.name);
		ti->list(&fi);
		return 0;
	}

	if (nimages == 0)
		usage(0);

	if (!opt.file && isatty(fileno(stdout)))
		errx(2, "Won't write to a tty (use -o)");

	for (l = images.next; l != &images; l = l->next) {
		struct image *im = list_item(l, struct image, list);
		read_im(im);
		if (im->has_key)
			add_transparency(im);
	}
	current = 0;

	out = stdout;
	if (opt.file) {
		out = fopen(opt.file, "wb");
		if (!out)
			err(1, "%s", opt.file);
	}

	write_ico(out);
	return 0;
}
